@@ -22,8 +22,10 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include "zix/hash.h"
-#include "zix/tree.h"
+#define ZIX_INLINE
+#include "zix/digest.c"
+#include "zix/hash.c"
+#include "zix/tree.c"
 
 #include "sord_config.h"
 #include "sord_internal.h"
@@ -102,10 +104,9 @@ static const int orderings[NUM_ORDERS][TUP_LEN] = {
 
 /** World */
 struct SordWorldImpl {
-	ZixHash* names;     ///< URI or blank node identifier string => ID
-	ZixHash* langs;     ///< Language tag => Interned language tag
-	ZixHash* literals;  ///< Literal => ID
-	size_t   n_nodes;   ///< Number of nodes
+	ZixHash*      nodes;
+	SerdErrorSink error_sink;
+	void*         error_handle;
 };
 
 /** Store */
@@ -141,66 +142,89 @@ struct SordIterImpl {
 	bool             skip_graphs;        ///< Iteration should ignore graphs
 };
 
-static unsigned
-sord_literal_hash(const void* n)
+static uint32_t
+sord_node_hash(const void* n)
 {
-	SordNode* node = (SordNode*)n;
-	return zix_string_hash(node->node.buf)
-		+ (node->lang ? zix_string_hash(node->lang) : 0);
+	const SordNode* node = (const SordNode*)n;
+	uint32_t        hash = zix_digest_start();
+	hash = zix_digest_add(hash, node->node.buf, node->node.n_bytes);
+	hash = zix_digest_add(hash, &node->node.type, sizeof(node->node.type));
+	if (node->node.type == SERD_LITERAL) {
+		hash = zix_digest_add(hash, &node->meta.lit, sizeof(node->meta.lit));
+	}
+	return hash;
 }
 
 static bool
-sord_literal_equal(const void* a, const void* b)
+sord_node_hash_equal(const void* a, const void* b)
 {
-	SordNode* a_node = (SordNode*)a;
-	SordNode* b_node = (SordNode*)b;
+	const SordNode* a_node = (const SordNode*)a;
+	const SordNode* b_node = (const SordNode*)b;
 	return (a_node == b_node)
-		|| (zix_string_equal(sord_node_get_string(a_node),
-		                     sord_node_get_string(b_node))
-		    && (a_node->lang == b_node->lang)
-		    && (a_node->datatype == b_node->datatype));
+		|| ((a_node->node.type == b_node->node.type) &&
+		    (a_node->node.type != SERD_LITERAL ||
+		     (a_node->meta.lit.datatype == b_node->meta.lit.datatype &&
+		      !strncmp(a_node->meta.lit.lang,
+		               b_node->meta.lit.lang,
+		               sizeof(a_node->meta.lit.lang)))) &&
+		    (serd_node_equals(&a_node->node, &b_node->node)));
+}
+
+static void
+error(SordWorld* world, SerdStatus st, const char* fmt, ...)
+{
+	va_list args;
+	va_start(args, fmt);
+	const SerdError e = { st, NULL, 0, 0, fmt, &args };
+	if (world->error_sink) {
+		world->error_sink(world->error_handle, &e);
+	} else {
+		fprintf(stderr, "error: ");
+		vfprintf(stderr, fmt, args);
+	}
+	va_end(args);
 }
 
 SordWorld*
 sord_world_new(void)
 {
 	SordWorld* world = (SordWorld*)malloc(sizeof(SordWorld));
-	world->names    = zix_hash_new(zix_string_hash, zix_string_equal);
-	world->langs    = zix_hash_new(zix_string_hash, zix_string_equal);
-	world->literals = zix_hash_new(sord_literal_hash, sord_literal_equal);
-	world->n_nodes  = 0;
+	world->error_sink   = NULL;
+	world->error_handle = NULL;
+
+	world->nodes = zix_hash_new(
+		sord_node_hash, sord_node_hash_equal, sizeof(SordNode));
+
 	return world;
 }
 
 static void
-free_node_entry(const void* key, void* value, void* user_data)
+free_node_entry(const void* value, void* user_data)
 {
-	SordNode* node = (SordNode*)value;
+	const SordNode* node = (const SordNode*)value;
 	if (node->node.type == SERD_LITERAL) {
-		sord_node_free((SordWorld*)user_data, node->datatype);
+		sord_node_free((SordWorld*)user_data, node->meta.lit.datatype);
 	}
 	free((uint8_t*)node->node.buf);
-	free(node);
-}
-
-static void
-free_lang_entry(const void* key, void* value, void* user_data)
-{
-	free(value);
 }
 
 void
 sord_world_free(SordWorld* world)
 {
-	zix_hash_foreach(world->literals, free_node_entry, world);
-	zix_hash_foreach(world->names, free_node_entry, world);
-	zix_hash_foreach(world->langs, free_lang_entry, world);
-	zix_hash_free(world->names);
-	zix_hash_free(world->langs);
-	zix_hash_free(world->literals);
+	zix_hash_foreach(world->nodes, free_node_entry, world);
+	zix_hash_free(world->nodes);
 	free(world);
 }
 
+void
+sord_world_set_error_sink(SordWorld*    world,
+                          SerdErrorSink error_sink,
+                          void*         handle)
+{
+	world->error_sink   = error_sink;
+	world->error_handle = handle;
+}
+
 /** Compare nodes, considering NULL a wildcard match. */
 static inline int
 sord_node_compare(const SordNode* a, const SordNode* b)
@@ -221,19 +245,15 @@ sord_node_compare(const SordNode* a, const SordNode* b)
 		             (const char*)sord_node_get_string(b));
 		if (cmp == 0) {
 			// Note: Can't use sord_node_compare here since it does wildcards
-			if (!a->datatype || !b->datatype) {
-				cmp = a->datatype - b->datatype;
+			if (!a->meta.lit.datatype || !b->meta.lit.datatype) {
+				cmp = a->meta.lit.datatype - b->meta.lit.datatype;
 			} else {
-				cmp = strcmp((const char*)a->datatype->node.buf,
-				             (const char*)b->datatype->node.buf);
+				cmp = strcmp((const char*)a->meta.lit.datatype->node.buf,
+				             (const char*)b->meta.lit.datatype->node.buf);
 			}
 		}
 		if (cmp == 0) {
-			if (!a->lang || !b->lang) {
-				cmp = a->lang - b->lang;
-			} else {
-				cmp = strcmp(a->lang, b->lang);
-			}
+			cmp = strcmp(a->meta.lit.lang, b->meta.lit.lang);
 		}
 	default:
 		break;
@@ -277,9 +297,9 @@ sord_quad_match(const SordQuad x, const SordQuad y)
 static int
 sord_quad_compare(const void* x_ptr, const void* y_ptr, void* user_data)
 {
-	const int* const ordering = (const int*)user_data;
-	SordNode** const x        = (SordNode**)x_ptr;
-	SordNode** const y        = (SordNode**)y_ptr;
+	const int* const           ordering = (const int*)user_data;
+	const SordNode*const*const x        = (const SordNode*const*)x_ptr;
+	const SordNode*const*const y        = (const SordNode*const*)y_ptr;
 
 	for (int i = 0; i < TUP_LEN; ++i) {
 		const int idx = ordering[i];
@@ -612,10 +632,10 @@ sord_new(SordWorld* world, unsigned indices, bool graphs)
 
 		if (indices & (1 << i)) {
 			sord->indices[i] = zix_tree_new(
-				false, sord_quad_compare, (void*)ordering);
+				false, sord_quad_compare, (void*)ordering, NULL);
 			if (graphs) {
 				sord->indices[i + (NUM_ORDERS / 2)] = zix_tree_new(
-					false, sord_quad_compare, (void*)g_ordering);
+					false, sord_quad_compare, (void*)g_ordering, NULL);
 			} else {
 				sord->indices[i + (NUM_ORDERS / 2)] = NULL;
 			}
@@ -627,11 +647,11 @@ sord_new(SordWorld* world, unsigned indices, bool graphs)
 
 	if (!sord->indices[DEFAULT_ORDER]) {
 		sord->indices[DEFAULT_ORDER] = zix_tree_new(
-			false, sord_quad_compare, (void*)orderings[DEFAULT_ORDER]);
+			false, sord_quad_compare, (void*)orderings[DEFAULT_ORDER], NULL);
 	}
 	if (graphs && !sord->indices[DEFAULT_GRAPH_ORDER]) {
 		sord->indices[DEFAULT_GRAPH_ORDER] = zix_tree_new(
-			false, sord_quad_compare, (void*)orderings[DEFAULT_GRAPH_ORDER]);
+			false, sord_quad_compare, (void*)orderings[DEFAULT_GRAPH_ORDER], NULL);
 	}
 
 	return sord;
@@ -641,20 +661,17 @@ static void
 sord_node_free_internal(SordWorld* world, SordNode* node)
 {
 	assert(node->refs == 0);
-	if (node->node.type == SERD_LITERAL) {
-		if (zix_hash_remove(world->literals, node)) {
-			fprintf(stderr, "Failed to remove literal from hash.\n");
-			return;
-		}
-		sord_node_free(world, node->datatype);
-	} else {
-		if (zix_hash_remove(world->names, node->node.buf)) {
-			fprintf(stderr, "Failed to remove resource from hash.\n");
-			return;
-		}
+
+	// Cache pointer to buffer to free after node removal and destruction
+	const uint8_t* const buf = node->node.buf;
+
+	// Remove node from hash (which frees the node)
+	if (zix_hash_remove(world->nodes, node)) {
+		error(world, SERD_ERR_INTERNAL, "failed to remove node from hash\n");
 	}
-	free((uint8_t*)node->node.buf);
-	free(node);
+
+	// Free buffer
+	free((uint8_t*)buf);
 }
 
 static void
@@ -663,8 +680,8 @@ sord_add_quad_ref(SordModel* sord, const SordNode* node, SordQuadIndex i)
 	if (node) {
 		assert(node->refs > 0);
 		++((SordNode*)node)->refs;
-		if (i == SORD_OBJECT) {
-			++((SordNode*)node)->refs_as_obj;
+		if (node->node.type != SERD_LITERAL && i == SORD_OBJECT) {
+			++((SordNode*)node)->meta.res.refs_as_obj;
 		}
 	}
 }
@@ -677,9 +694,9 @@ sord_drop_quad_ref(SordModel* sord, const SordNode* node, SordQuadIndex i)
 	}
 
 	assert(node->refs > 0);
-	if (i == SORD_OBJECT) {
-		assert(node->refs_as_obj > 0);
-		--((SordNode*)node)->refs_as_obj;
+	if (node->node.type != SERD_LITERAL && i == SORD_OBJECT) {
+		assert(node->meta.res.refs_as_obj > 0);
+		--((SordNode*)node)->meta.res.refs_as_obj;
 	}
 	if (--((SordNode*)node)->refs == 0) {
 		sord_node_free_internal(sord_get_world(sord), (SordNode*)node);
@@ -697,23 +714,23 @@ sord_free(SordModel* sord)
 	SordIter* i = sord_begin(sord);
 	for (; !sord_iter_end(i); sord_iter_next(i)) {
 		sord_iter_get(i, tup);
-		for (int i = 0; i < TUP_LEN; ++i) {
-			sord_drop_quad_ref(sord, (SordNode*)tup[i], (SordQuadIndex)i);
+		for (int t = 0; t < TUP_LEN; ++t) {
+			sord_drop_quad_ref(sord, tup[t], (SordQuadIndex)t);
 		}
 	}
 	sord_iter_free(i);
 
 	// Free quads
-	for (ZixTreeIter* i = zix_tree_begin(sord->indices[DEFAULT_ORDER]);
-	     !zix_tree_iter_is_end(i);
-	     i = zix_tree_iter_next(i)) {
-		free(zix_tree_get(i));
+	for (ZixTreeIter* t = zix_tree_begin(sord->indices[DEFAULT_ORDER]);
+	     !zix_tree_iter_is_end(t);
+	     t = zix_tree_iter_next(t)) {
+		free(zix_tree_get(t));
 	}
 
 	// Free indices
-	for (unsigned i = 0; i < NUM_ORDERS; ++i)
-		if (sord->indices[i])
-			zix_tree_free(sord->indices[i]);
+	for (unsigned o = 0; o < NUM_ORDERS; ++o)
+		if (sord->indices[o])
+			zix_tree_free(sord->indices[o]);
 
 	free(sord);
 }
@@ -733,7 +750,7 @@ sord_num_quads(const SordModel* sord)
 size_t
 sord_num_nodes(const SordWorld* world)
 {
-	return world->n_nodes;
+	return zix_hash_size(world->nodes);
 }
 
 SordIter*
@@ -752,7 +769,7 @@ static inline ZixTreeIter*
 index_search(ZixTree* db, const SordQuad search_key)
 {
 	ZixTreeIter* iter = NULL;
-	zix_tree_find(db, (void*)search_key, &iter);
+	zix_tree_find(db, (const void*)search_key, &iter);
 	return iter;
 }
 
@@ -760,7 +777,7 @@ static inline ZixTreeIter*
 index_lower_bound(ZixTree* db, const SordQuad search_key)
 {
 	ZixTreeIter* iter = NULL;
-	zix_tree_find(db, (void*)search_key, &iter);
+	zix_tree_find(db, (const void*)search_key, &iter);
 	if (!iter) {
 		return NULL;
 	}
@@ -863,75 +880,14 @@ sord_contains(SordModel* sord, const SordQuad pat)
 	return ret;
 }
 
-static SordNode*
-sord_lookup_name(SordWorld* world, const uint8_t* str)
+static uint8_t*
+sord_strndup(const uint8_t* str, size_t len)
 {
-	return (SordNode*)zix_hash_find(world->names, str);
-}
-
-char*
-sord_strndup(const char* str, size_t len)
-{
-	char* dup = (char*)malloc(len + 1);
+	uint8_t* dup = (uint8_t*)malloc(len + 1);
 	memcpy(dup, str, len + 1);
 	return dup;
 }
 
-static SordNode*
-sord_new_node(SerdType type, const uint8_t* data,
-              size_t n_bytes, size_t n_chars, SerdNodeFlags flags,
-              SordNode* datatype, const char* lang, bool copy)
-{
-	SordNode* node = (SordNode*)malloc(sizeof(SordNode));
-	node->lang         = lang;
-	node->datatype     = datatype;
-	node->refs         = 1;
-	node->refs_as_obj  = 0;
-	node->node.buf     = (uint8_t*)data;
-	node->node.n_bytes = n_bytes;
-	node->node.n_chars = n_chars;
-	node->node.flags   = flags;
-	node->node.type    = type;
-
-	if (copy) {
-		node->node.buf = (uint8_t*)sord_strndup((const char*)data, n_bytes);
-	}
-	return node;
-}
-
-const char*
-sord_intern_lang(SordWorld* world, const char* lang)
-{
-	if (lang) {
-		char* ilang = (char*)zix_hash_find(world->langs, lang);
-		if (!ilang) {
-			ilang = sord_strndup(lang, strlen(lang));
-			zix_hash_insert(world->langs, ilang, ilang);
-		}
-		lang = ilang;
-	}
-	return lang;
-}
-
-static SordNode*
-sord_lookup_literal(SordWorld* world, SordNode* type,
-                    const uint8_t* str, size_t n_bytes, size_t n_chars,
-                    const char*    lang)
-{
-	struct SordNodeImpl key;
-	key.lang         = lang;
-	key.datatype     = type;
-	key.refs         = 1;
-	key.refs_as_obj  = 1;
-	key.node.buf     = (uint8_t*)str;
-	key.node.n_bytes = n_bytes;
-	key.node.n_chars = n_chars;
-	key.node.flags   = 0;
-	key.node.type    = SERD_LITERAL;
-
-	return (SordNode*)zix_hash_find(world->literals, &key);
-}
-
 SordNodeType
 sord_node_get_type(const SordNode* node)
 {
@@ -943,34 +899,37 @@ sord_node_get_type(const SordNode* node)
 	case SERD_URI:
 		return SORD_URI;
 	default:
-		fprintf(stderr, "sord: error: Illegal node type.\n");
+		fprintf(stderr, "error: invalid node type\n");
 		return (SordNodeType)0;
 	}
 }
 
 const uint8_t*
-sord_node_get_string(const SordNode* ref)
+sord_node_get_string(const SordNode* node)
 {
-	return ref->node.buf;
+	return node->node.buf;
 }
 
 const uint8_t*
-sord_node_get_string_counted(const SordNode* ref, size_t* len)
+sord_node_get_string_counted(const SordNode* node, size_t* len)
 {
-	*len = ref->node.n_chars;
-	return ref->node.buf;
+	*len = node->node.n_chars;
+	return node->node.buf;
 }
 
 const char*
-sord_node_get_language(const SordNode* ref)
+sord_node_get_language(const SordNode* node)
 {
-	return ref->lang;
+	if (node->node.type != SERD_LITERAL || !node->meta.lit.lang[0]) {
+		return NULL;
+	}
+	return node->meta.lit.lang;
 }
 
 SordNode*
-sord_node_get_datatype(const SordNode* ref)
+sord_node_get_datatype(const SordNode* node)
 {
-	return ref->datatype;
+	return (node->node.type == SERD_LITERAL) ? node->meta.lit.datatype : NULL;
 }
 
 SerdNodeFlags
@@ -982,13 +941,39 @@ sord_node_get_flags(const SordNode* node)
 bool
 sord_node_is_inline_object(const SordNode* node)
 {
-	return (node->node.type == SERD_BLANK) && (node->refs_as_obj == 1);
+	return (node->node.type == SERD_BLANK) && (node->meta.res.refs_as_obj == 1);
 }
 
-static void
-sord_add_node(SordWorld* world, SordNode* node)
+static SordNode*
+sord_insert_node(SordWorld* world, const SordNode* key, bool copy)
 {
-	++world->n_nodes;
+	SordNode* node = NULL;
+	ZixStatus st   = zix_hash_insert(world->nodes, key, (const void**)&node);
+	switch (st) {
+	case ZIX_STATUS_EXISTS:
+		++node->refs;
+		break;
+	case ZIX_STATUS_SUCCESS:
+		assert(node->refs == 1);
+		if (copy) {
+			node->node.buf = sord_strndup(node->node.buf, node->node.n_bytes);
+		}
+		if (node->node.type == SERD_LITERAL) {
+			node->meta.lit.datatype = sord_node_copy(node->meta.lit.datatype);
+		}
+		return node;
+	default:
+		assert(!node);
+		error(world, SERD_ERR_INTERNAL,
+		      "error inserting node `%s'\n", key->node.buf);
+	}
+
+	if (!copy) {
+		// Free the buffer we would have copied if a new node was created
+		free((uint8_t*)key->node.buf);
+	}
+
+	return node;
 }
 
 static SordNode*
@@ -996,24 +981,16 @@ sord_new_uri_counted(SordWorld* world, const uint8_t* str,
                      size_t n_bytes, size_t n_chars, bool copy)
 {
 	if (!serd_uri_string_has_scheme(str)) {
-		fprintf(stderr, "Attempt to map invalid URI `%s'.\n", str);
+		error(world, SERD_ERR_BAD_ARG,
+		      "attempt to map invalid URI `%s'\n", str);
 		return NULL;  // Can't intern relative URIs
 	}
 
-	SordNode* node = sord_lookup_name(world, str);
-	if (node) {
-		if (!copy) {
-			free((uint8_t*)str);
-		}
-		++node->refs;
-		return node;
-	}
+	const SordNode key = {
+		{ str, n_bytes, n_chars, 0, SERD_URI }, 1, { { 0 } }
+	};
 
-	node = sord_new_node(SERD_URI, str, n_bytes, n_chars, 0, 0, 0, copy);
-	assert(!zix_hash_find(world->names, node->node.buf));
-	zix_hash_insert(world->names, (char*)node->node.buf, node);
-	sord_add_node(world, node);
-	return node;
+	return sord_insert_node(world, &key, copy);
 }
 
 SordNode*
@@ -1046,16 +1023,11 @@ static SordNode*
 sord_new_blank_counted(SordWorld* world, const uint8_t* str,
                        size_t n_bytes, size_t n_chars)
 {
-	SordNode* node = sord_lookup_name(world, str);
-	if (node) {
-		++node->refs;
-		return node;
-	}
+	const SordNode key = {
+		{ str, n_bytes, n_chars, 0, SERD_BLANK }, 1, { { 0 } }
+	};
 
-	node = sord_new_node(SERD_BLANK, str, n_bytes, n_chars, 0, 0, 0, true);
-	zix_hash_insert(world->names, (char*)node->node.buf, node);
-	sord_add_node(world, node);
-	return node;
+	return sord_insert_node(world, &key, true);
 }
 
 SordNode*
@@ -1074,21 +1046,16 @@ sord_new_literal_counted(SordWorld*     world,
                          SerdNodeFlags  flags,
                          const char*    lang)
 {
-	lang = sord_intern_lang(world, lang);
-	SordNode* node = sord_lookup_literal(
-		world, datatype, str, n_bytes, n_chars, lang);
-	if (node) {
-		++node->refs;
-		return node;
+	SordNode key = {
+		{ str, n_bytes, n_chars, flags, SERD_LITERAL }, 1, { { 0 } }
+	};
+	key.meta.lit.datatype = datatype;
+	memset(key.meta.lit.lang, 0, sizeof(key.meta.lit.lang));
+	if (lang) {
+		strncpy(key.meta.lit.lang, lang, sizeof(key.meta.lit.lang));
 	}
 
-	node = sord_new_node(SERD_LITERAL,
-	                     str, n_bytes, n_chars, flags,
-	                     sord_node_copy(datatype), lang, true);
-	zix_hash_insert(world->literals, node, node);  // FIXME: correct?
-	sord_add_node(world, node);
-	assert(node->refs == 1);
-	return node;
+	return sord_insert_node(world, &key, true);
 }
 
 SordNode*
@@ -1142,11 +1109,11 @@ sord_node_from_serd_node(SordWorld*      world,
 			SerdURI  abs_uri;
 			SerdNode abs_uri_node = serd_node_new_uri_from_node(
 				sn, &base_uri, &abs_uri);
-			SordNode* ret = sord_new_uri_counted(world,
-			                                     abs_uri_node.buf,
-			                                     abs_uri_node.n_bytes,
-			                                     abs_uri_node.n_chars,
-			                                     true);
+			ret = sord_new_uri_counted(world,
+			                           abs_uri_node.buf,
+			                           abs_uri_node.n_bytes,
+			                           abs_uri_node.n_chars,
+			                           true);
 			serd_node_free(&abs_uri_node);
 			return ret;
 		}
@@ -1154,7 +1121,8 @@ sord_node_from_serd_node(SordWorld*      world,
 		SerdChunk uri_prefix;
 		SerdChunk uri_suffix;
 		if (serd_env_expand(env, sn, &uri_prefix, &uri_suffix)) {
-			fprintf(stderr, "Failed to expand qname `%s'\n", sn->buf);
+			error(world, SERD_ERR_BAD_CURIE,
+			      "failed to expand CURIE `%s'\n", sn->buf);
 			return NULL;
 		}
 		const size_t uri_len = uri_prefix.len + uri_suffix.len;
@@ -1162,7 +1130,7 @@ sord_node_from_serd_node(SordWorld*      world,
 		memcpy(buf,                  uri_prefix.buf, uri_prefix.len);
 		memcpy(buf + uri_prefix.len, uri_suffix.buf, uri_suffix.len);
 		buf[uri_len] = '\0';
-		SordNode* ret = sord_new_uri_counted(
+		ret = sord_new_uri_counted(
 			world, buf, uri_prefix.len + uri_suffix.len,
 			uri_prefix.len + uri_suffix.len, false);  // FIXME: UTF-8
 		return ret;
@@ -1213,7 +1181,8 @@ sord_add(SordModel* sord, const SordQuad tup)
 {
 	SORD_WRITE_LOG("Add " TUP_FMT "\n", TUP_FMT_ARGS(tup));
 	if (!tup[0] || !tup[1] || !tup[2]) {
-		fprintf(stderr, "Attempt to add quad with NULL field.\n");
+		error(sord->world, SERD_ERR_BAD_ARG,
+		      "attempt to add quad with NULL field\n");
 		return false;
 	}
 
